import logging
import os
from abc import ABC, abstractmethod
from types import SimpleNamespace

import mysql.connector as mysql
import redis
import azure.cosmosdb.table

from .exception import PipelineError

FORMAT = '%(asctime)-15s %(levelname)s %(message)s'
logging.basicConfig(format=FORMAT)
logger = logging.getLogger('worker')
logger.setLevel(logging.DEBUG)


def KindsOfCache():
    return [cls.kind for cls in Cache.__subclasses__()]


def CacheOf(typename):
    for cls in Cache.__subclasses__():
        if cls.is_cls_of(typename):
            return cls


def parse_connection_string(connectionString, no_port=False, no_username=False, defaults={}):
    """ Parse connection string in user:password@host:port format

    >>> parse_connection_string("username:password@host:port")
    namespace(host='host', password='password', port='port', username='username')
    >>> parse_connection_string("username@host")
    namespace(host='host', password=None, port=None, username='username')
    >>> parse_connection_string("username:password@host:port", no_port=True)
    namespace(host='host:port', password='password', port=None, username='username')
    >>> parse_connection_string("password@host:port", no_username=True)
    namespace(host='host', password='password', port='port', username=None)
    """
    *userNameAndPasswordOrEmpty, remaining = connectionString.split('@')
    password = None
    if userNameAndPasswordOrEmpty:
        if no_username:
            username, password = None, userNameAndPasswordOrEmpty[0]
        else:
            username, *passwordOrEmpty = userNameAndPasswordOrEmpty[0].split(':')
            if passwordOrEmpty:
                password = passwordOrEmpty[0]
    else:
        username = None
    if not username:
        username = defaults.get('username')

    port = None
    if no_port:
        host = remaining
    else:
        host, *portOrEmpty = remaining.split(':')
        if portOrEmpty:
            port = portOrEmpty[0]
    if not port:
        port = defaults.get('port')

    return SimpleNamespace(
        host=host,
        port=port,
        username=username,
        password=password,
    )


class Cache(ABC):
    kind = 'NONE'

    def __init__(self, config, logger=logger):
        self.config = config
        self.in_fields = self.config.in_fields.split(',') if self.config.in_fields else []
        self.out_fields = self.config.out_fields.split(',') if self.config.out_fields else []
        self.logger = logger

    @classmethod
    def add_arguments(cls, parser):
        parser.add_argument('--in-fields', type=str,
                            default=os.environ.get('INFIELDS', None),
                            help='database fields to read (comma separated)')
        parser.add_argument('--out-fields', type=str,
                            default=os.environ.get('OUTFIELDS', None),
                            help='database fields to write (comma separated)')

    def setup(self):
        """ establish connection for example """

    @abstractmethod
    def read(self, key):
        """ return corresponding *fields* from database with key """

    @abstractmethod
    def write(self, key, kv):
        """ return corresponding *fields* from database with key """

    @classmethod
    def is_cls_of(cls, kind):
        return kind == cls.kind

    def close(self):
        pass


class MemoryCache(Cache):
    """ MemoryReader read data from memory.
    It is for testing only.

    >>> from types import SimpleNamespace
    >>> data = {'key1':{'text':'text1', 'title':'title1'}}
    >>> mem = MemoryCache(SimpleNamespace(in_fields='title', out_fields='text,title', mem=data))
    >>> mem.write('key3', {'text': 'text3', 'title': 'title3'})
    >>> mem.read('key3')
    {'title': 'title3'}
    """
    kind = 'MEM'

    def __repr__(self):
        return 'MemoryCache'

    def read(self, key):
        dct = self.config.mem[key]
        if self.in_fields:
            dct = dict([(k, dct[k]) for k in self.in_fields])
        return dct

    def write(self, key, kvs):
        self.config.mem[key] = kvs


class MySqlCache(Cache):
    """ MySQLReader reads data from MySQL

    >>> from unittest.mock import patch
    >>> from argparse import ArgumentParser
    >>> parser = ArgumentParser()
    >>> MySqlCache.add_arguments(parser)
    >>> config = parser.parse_args(["--in-fields", "text"])
    >>> with patch('mysql.connector.connect') as c:
    ...     MySqlCache(config, "text")
    MySqlCache(localhost:3306/database/table):['text']:[]
    """
    kind = 'MYSQL'

    def __init__(self, config, logger=logger):
        super().__init__(config, logger)
        self.setup()

    def __repr__(self):
        return 'MySqlCache({}/{}/{}):{}:{}'.format(
            self.mysqlConfig.host,
            self.config.database,
            self.config.table,
            self.in_fields,
            self.out_fields,
        )

    @classmethod
    def add_arguments(cls, parser):
        super().add_arguments(parser)
        parser.add_argument(
            '--mysql', type=str,
            default=os.environ.get('MYSQL', 'localhost:3306'),
            help='mysql database host:port'
        )
        parser.add_argument(
            '--database', type=str,
            default=os.environ.get('MYSQLDATABASE', 'database'),
            help='muysql database'
        )
        parser.add_argument(
            '--table', type=str,
            default=os.environ.get('MYSQLTABLE', 'table'),
            help='mysql database table'
        )
        parser.add_argument(
            '--keyname', type=str,
            default=os.environ.get('KEYNAME', 'id'),
            help='database field name of key'
        )

    def setup(self):
        self.mysqlConfig = parse_connection_string(self.config.mysql, no_port=True)
        self.db = mysql.connect(
            host=self.mysqlConfig.host,
            user=self.mysqlConfig.username,
            passwd=self.mysqlConfig.password,
            database=self.config.database,
        )
        self.cursor = self.db.cursor()

    def read(self, key):
        query = "SELECT {} FROM {} WHERE {} = '{}'".format(
            self.config.in_fields,
            self.config.table,
            self.config.keyname,
            key
        )
        self.cursor.execute(query)
        r = self.cursor.fetchone()
        if r is None:
            raise PipelineError('Data with key is not found')
        return dict(zip(
            self.config.in_fields.split(','),
            r
        ))

    def write(self, key, kvs):
        query = "UPDATE {} SET {} WHERE {} = '{}'".format(
            self.config.table,
            ",".join(["{} = {}".format(k, v) for k, v in kvs.items()]),
            self.config.keyname,
            key
        )
        self.cursor.execute(query)
        self.db.commit()


class RedisCache(Cache):
    """ RedisCache reads/writes data from/to Redis

    >>> from unittest.mock import patch
    >>> from argparse import ArgumentParser
    >>> parser = ArgumentParser()
    >>> RedisCache.add_arguments(parser)
    >>> config = parser.parse_args(["--in-fields", "text,title"])
    >>> with patch('redis.Redis') as c:
    ...     RedisCache(config)
    RedisCache(localhost:6379):['text', 'title']:[]
    """
    kind = 'REDIS'

    def __init__(self, config, logger=logger):
        super().__init__(config, logger)
        self.setup()

    def __repr__(self):
        return 'RedisCache({}:{}):{}:{}'.format(
            self.redisConfig.host,
            self.redisConfig.port,
            self.in_fields,
            self.out_fields,
        )

    @classmethod
    def add_arguments(cls, parser):
        super().add_arguments(parser)
        parser.add_argument(
            '--redis', type=str,
            default=os.environ.get('REDIS', 'localhost:6379'),
            help='redis host:port'
        )
        parser.add_argument(
            '--expire', type=int,
            default=os.environ.get('REDISEXPIRE', 7*86400),
            help='expire time for database (default: 7 days)'
        )

    def setup(self):
        self.redisConfig = parse_connection_string(self.config.redis, no_username=True)
        self.redis = redis.Redis(
            host=self.redisConfig.host,
            port=self.redisConfig.port,
            password=self.redisConfig.password,
        )

    def read(self, key):
        """ entries are stored as following in redis:
            a set is managed for each key to contain fields available
            a key:field -> value for accessing field for each key

            TODO: raise error if fields are not available
        """
        results = self.redis.mget(['{}:{}'.format(key, field) for field in self.in_fields])
        return dict(zip(self.fields, results))

    def write(self, key, kvs):
        """ entries are stored as following in redis:
            a set is managed for each key to contain fields available
            a key:field -> value for accessing field for each key

            TODO: check error after mset
        """
        self.redis.mset(
            dict([(key, k) for k, v in kvs.items()]),
        )
        for key in kvs.keys():
            self.redis.expire(key, self.config.expire)


class AzureTableCache(Cache):
    """ AzureTableCache reads/writes data from/to Azure Table

    >>> from unittest.mock import patch
    >>> from argparse import ArgumentParser
    >>> parser = ArgumentParser()
    >>> AzureTableCache.add_arguments(parser)
    >>> config = parser.parse_args(["--in-fields", "text,title"])
    >>> with patch('azure.cosmosdb.table.TableService') as c:
    ...     AzureTableCache(config)
    AzureTableCache:['text', 'title']:[]
    """
    kind = 'AZURE'

    def __init__(self, config, logger=logger):
        super().__init__(config, logger)
        self.setup()

    def __repr__(self):
        return 'AzureTableCache:{}:{}'.format(self.in_fields, self.out_fields)

    @classmethod
    def add_arguments(cls, parser):
        super().add_arguments(parser)
        parser.add_argument(
            '--azuredb', type=str,
            default=os.environ.get('AZUREDB', ''),
            help='azure table connection string'
        )
        parser.add_argument(
            '--table', type=str,
            default=os.environ.get('TABLE', ''),
            help='azure table table name'
        )

    def setup(self):
        self.service = azure.cosmosdb.table.TableService(
            endpoint_suffix="table.cosmos.azure.com",
            connection_string=self.config.azuredb
        )

    def read(self, key):
        """ entries are stored as following in redis:
            a set is managed for each key to contain fields available
            a key:field -> value for accessing field for each key

            TODO: raise error if fields are not available
        """
        entity = self.service.get_entity(
            self.config.table,
            'key',  # PartitionKey
            key,  # RowKey
            timeout=5.0,
        )
        if self.in_fields:
            fields = self.in_fields
        else:
            fields = [k for k in entity.keys() if k not in ('PartitionKey', 'RowKey', 'Timestamp', 'etag')]
        return {k: v for k, v in entity.items() if k in fields}

    def write(self, key, kvs):
        """ entries are stored as following in redis:
            a set is managed for each key to contain fields available
            a key:field -> value for accessing field for each key
        """
        dct = {
            'PartitionKey': 'key',
            'RowKey': key,
        }
        dct.update(kvs)
        self.service.insert_or_merge_entity(
            self.config.table,
            dct,
            timeout=5.0,
        )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os
import traceback

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.doctest',
    'sphinx.ext.extlinks',
    'sphinx.ext.ifconfig',
    'sphinx.ext.napoleon',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
]
source_suffix = '.rst'
master_doc = 'index'
project = 'Tanbih Pipeline'
year = '2020'
author = 'Yifan Zhang'
copyright = '{0}, {1}'.format(year, author)
release = '0.4.2'
#try:
#    from pkg_resources import get_distribution
#    version = release = get_distribution('tanbih-worker').version
#except Exception:
#    traceback.print_exc()
#    version = release = '0.4.2'

pygments_style = 'trac'
templates_path = ['.']
extlinks = {
    'issue': ('https://github.com/yifan/pipeline/issues/%s', '#'),
    'pr': ('https://github.com/yifan/pipeline/pulls/%s', 'PR #'),
}
#import sphinx_py3doc_enhanced_theme
#html_theme = "sphinx_py3doc_enhanced_theme"
#html_theme_path = [sphinx_py3doc_enhanced_theme.get_html_theme_path()]
#html_theme_options = {
#    'githuburl': 'https://gitlab.com/meganews/pulsarworker'
#}
#
#html_use_smartypants = True
#html_last_updated_fmt = '%b %d, %Y'
#html_split_index = False
#html_sidebars = {
#   '**': ['searchbox.html', 'globaltoc.html', 'sourcelink.html'],
#}
#html_short_title = '%s-%s' % (project, version)

napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = False



import logging
from unittest import TestCase

from pipeline import PipelineError


class TestException(TestCase):
    def test_default(self):
        try:
            raise PipelineError("Test")
        except PipelineError as e:
            e.log(logging.getLogger('test'))



class PipelineError(Exception):
    def __init__(self, message, data=None, traceback=None):
        super().__init__(message)
        self.data = data
        self.traceback = traceback

    def log(self, logger):
        if self.data:
            if hasattr(self.data, 'log'):
                self.data(logger)
            else:
                logger.warn(self.data)
        if self.traceback:
            logger.warning(self.traceback)



from unittest import TestCase

from pipeline import DestinationTap, SourceTap

BASE_DIR = './'


class DummySourceTap(SourceTap):
    kind = 'DUMMY'

    def __next__(self):
        pass

    def rewind(self):
        pass


class DummyDestinationTap(DestinationTap):
    kind = 'DUMMY'

    def send(self, msg):
        pass


class TestTaps(TestCase):
    def test_is_cls_of(self):
        self.assertTrue(DummySourceTap.is_cls_of('DUMMY'))
        self.assertTrue(DummyDestinationTap.is_cls_of('DUMMY'))



import argparse
import logging
import os
import sys
import traceback
from abc import ABC
from copy import copy

from .cache import CacheOf
from .exception import PipelineError
from .message import Message
from .monitor import Monitor
from .tap import DestinationOf, SourceOf
from .utils import parse_kind

logger = logging.getLogger('pipeline')
logger.setLevel(logging.INFO)


class WorkerConfig:
    NO_FLAG = 0
    NO_INPUT = 1
    NO_OUTPUT = 2

    def __init__(self, noInput=False, noOutput=False, messageClass=Message, cacheKind=None):
        if noInput and noOutput:
            raise ValueError('Worker cannot has no input AND no output')
        self.flag = WorkerConfig.NO_FLAG
        if noInput:
            self.flag = WorkerConfig.NO_INPUT
        if noOutput:
            self.flag = WorkerConfig.NO_OUTPUT
        self.messageClass = messageClass
        self.cacheKind = cacheKind

    def disable_input(self):
        self.flag = WorkerConfig.NO_INPUT

    def disable_output(self):
        self.flag = WorkerConfig.NO_OUTPUT


class WorkerCore(ABC):
    """Internal base class for pulsar-worker, DO NOT use it in your program!"""

    def __init__(self, name, version, description, config):
        self.name = name
        self.version = [int(x) for x in version.split('.')]
        if len(self.version) != 3:
            raise RuntimeError('Version format is not valid [x.x.x]')
        self.description = description
        self.kind = None
        self.logger = logger
        self.flag = config.flag
        self.messageClass = config.messageClass
        self.cacheKind = config.cacheKind

        self.parser = argparse.ArgumentParser(
            prog=name,
            description=description,
            conflict_handler='resolve',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter
        )

        self.monitor = Monitor(self)

        self.logger.info("Pipeline Worker %s (%s)", name, version)

    def setup(self):
        """loading code goes here"""

    def has_no_input(self):
        return self.flag == WorkerConfig.NO_INPUT

    def has_no_output(self):
        return self.flag == WorkerConfig.NO_OUTPUT

    def parse_args(self, args=sys.argv[1:], config=None):
        known, extras = parse_kind(args)
        self.kind = known.kind
        self.cacheKind = known.cacheKind or self.cacheKind
        if self.flag != WorkerConfig.NO_INPUT:
            self.sourceClass = SourceOf(self.kind)
        if self.flag != WorkerConfig.NO_OUTPUT:
            self.destinationClass = DestinationOf(self.kind)
        if self.cacheKind:
            self.cacheClass = CacheOf(self.cacheKind)
        self._add_arguments(self.parser)
        if config:
            self.parser.set_defaults(**config)
        self.parser.set_defaults(message=self.messageClass)
        self.options = self.parser.parse_args(extras)
        if self.flag != WorkerConfig.NO_INPUT:
            self.source = self.sourceClass(self.options, logger=self.logger)
        if self.flag != WorkerConfig.NO_OUTPUT:
            self.destination = self.destinationClass(self.options, logger=self.logger)
        if self.cacheKind:
            self.cache = self.cacheClass(self.options, logger=self.logger)
        # report worker info to monitor
        self.monitor.record_worker_info()

    def _add_arguments(self, parser):
        parser.add_argument('--debug', action='store_true', default=os.environ.get('DEBUG', 'FALSE') == 'TRUE',
                            help='debug, more verbose logging')
        if self.flag != WorkerConfig.NO_INPUT:
            self.sourceClass.add_arguments(parser)
        if self.flag != WorkerConfig.NO_OUTPUT:
            self.destinationClass.add_arguments(parser)
        if self.cacheKind:
            self.cacheClass.add_arguments(parser)
        self.add_arguments(parser)

    def add_arguments(self, parser):
        """Add commandline arguments, to be override by child classes."""


class GeneratorConfig(WorkerConfig):
    pass


class Generator(WorkerCore):
    def __init__(self, name, version, description=None, config=GeneratorConfig()):
        config.disable_input()
        super().__init__(name, version, description, config)
        self.generator = None

    def generate(self):
        """a generator to generate dict."""
        yield self.messageClass()

    def step(self):
        if not self.generator:
            self.generator = self.generate()
        dct = next(self.generator)
        msg = self.messageClass(dct)
        msg.update_version(self.name, self.version)
        self.logger.info('Generated %s', msg)
        if msg.is_valid():
            self.destination.write(msg)
            self.monitor.record_write(self.destination.topic)
        else:
            self.logger.error('Generated message is invalid, skipping')
            self.logger.error(msg.log_info())
            self.logger.error(msg.log_content())
            raise PipelineError('Invalid message')

    def start(self, monitoring=False):
        try:
            options = self.options
        except AttributeError:
            logger.critical('Did you forget to run .parse_args before start?')
            raise

        self.logger.setLevel(level=logging.INFO)
        if options.debug:
            self.logger.setLevel(level=logging.DEBUG)

        self.logger.info('settings: write topic %s', options.out_topic)

        self.setup()

        if monitoring:
            self.monitor.expose()

        self.monitor.record_start()

        try:
            while True:
                self.step()
        except StopIteration:
            pass
        except PipelineError as e:
            e.log(self.logger)
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.monitor.record_error(str(e))
        self.destination.close()

        self.monitor.record_finish()


class SplitterConfig(WorkerConfig):
    pass


class Splitter(WorkerCore):
    """ Splitter will write to a topic whose name is based on a function
    """
    def __init__(self, name, version, description=None, config=SplitterConfig()):
        super().__init__(name, version, description, config)
        # keep a dictionary for 'topic': 'destination', self.destination is only used to parse
        # command line arguments
        self.destinations = {}

    def get_topic(self, msg):
        return '{}-{}'.format(self.destination.topic, msg.get('language', ''))

    def _run_streaming(self):
        for msg in self.source.read():
            self.logger.info("Received message '%s'", str(msg))
            self.monitor.record_read(self.source.topic)

            topic = self.get_topic(msg)

            if topic not in self.destinations:
                config = copy(self.destination.config)
                config.out_topic = topic
                self.destinations[topic] = self.destinationClass(config, logger=self.logger)

            destination = self.destinations[topic]

            if msg.is_valid():
                destination.write(msg)
            else:
                self.logger.error('Produced message is invalid, skipping')
                self.logger.error(msg.log_info())
                self.logger.error(msg.log_content())
            self.monitor.record_write(topic)
            self.source.acknowledge()

    def start(self, monitoring=False):
        try:
            options = self.options
        except AttributeError:
            self.logger.critical('Did you forget to run .parse_args before start?')
            raise

        self.logger.setLevel(level=logging.INFO)
        if options.debug:
            self.logger.setLevel(level=logging.DEBUG)

        if options.rewind:
            # consumer.seek(pulsar.MessageId.earliest)
            self.logger.info('seeked to earliest message available as requested')

        self.setup()

        self.logger.info('start listening')
        # if batch_mode:
        #  self._run_batch()
        # else:

        if monitoring:
            self.monitor.expose()

        self.monitor.record_start()
        try:
            self._run_streaming()
            self.source.close()
            self.destination.close()
        except PipelineError as e:
            e.log(self.logger)
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.monitor.record_error(str(e))
        self.monitor.record_finish()


class ProcessorConfig(WorkerConfig):
    pass


class Processor(WorkerCore):
    def __init__(self, name, version, description=None, config=ProcessorConfig()):
        super().__init__(name, version, description, config)
        self.retryEnabled = False

    def use_retry_topic(self, name=None):
        """ Retry topic is introduced to solve error handling by saving
            message being processed when error occurs to a separate topic.
            In this way, these messages can be reprocessed at a later stage.
        """
        config = copy(self.source.config)
        config.out_topic = name if name else self.name + "-retry"
        self.retryDestination = self.destinationClass(config, logger=self.logger)
        self.retryEnabled = True

    def process(self, msg):
        """ process function to be overridden by users, for streaming
            processing, this function needs to do in-place update on msg.dct
            and return an error or a list of errors (for batch processing).
            Message has been terminated though .terminates() will be skipped
            in output.
            A typical process definition will be:
                value = msg.get('preExistingKey')
                updates = {
                    'newKey': 'newValue',
                }
                msg.update(updates)
        """
        return None

    def _step(self, msg):
        """ process one message """
        failedOnError = False

        # process message
        if msg.should_update(self.name, self.version):
            self.logger.info("Processing message '%s'", str(msg))

            try:
                err = self.process(msg)
            except Exception:
                self.logger.error(traceback.format_exc())
                self.logger.error(msg.log_content())
                raise

            if err:
                self.logger.error("Error has occurred for message '%s': %s", str(msg), err)
                failedOnError = True
            else:
                msg.update_version(self.name, self.version)
        else:
            self.logger.warn('Message has been processed by higher version processor, no processed')

        # skip validation and output for some cases
        if msg.terminated:
            self.logger.info('Message<%s> terminates here', str(msg))
        elif not any((failedOnError, self.has_no_output())):
            if msg.is_valid():
                self.logger.info("Writing message '%s'", str(msg))
                self.destination.write(msg)
                self.monitor.record_write(self.destination.topic)
            else:
                failedOnError = True
                self.logger.error('result message is invalid, skipping')
                self.logger.warn(msg.log_info())
                self.logger.warn(msg.log_content())

        # retry if necessary
        if failedOnError and self.retryEnabled:
            self.logger.warn('message is sent to retry topic %s', self.retryDestination.config.out_topic)
            self.retryDestination.write(msg)
            self.monitor.record_write(self.retryDestination.topic)

    def _run_streaming(self):
        """ streaming processing messages from source and write resulted messages to destination

            error handling:
            1) error indicated by .process returning non-None value
                message will be written to retry topic if retry topic is enabled.
                WARNING: The message may have been modified during process().
            2) the result message is invalid
                the result message will be written to retry topic if retry topic is enabled.
                The message is processed, and invalid.
                WARNING: Processor will not re-process message unless the version of processor
                is higher than the version stored in message info.
        """
        for msg in self.source.read():
            self.monitor.record_read(self.source.topic)
            self.logger.info("Received message '%s'", str(msg))

            self._step(msg)

            self.source.acknowledge()

    def start(self, batch_mode=False, monitoring=False):
        """ start processing. """
        try:
            options = self.options
        except AttributeError:
            self.logger.critical('Did you forget to run .parse_args before start?')
            raise

        self.logger.setLevel(level=logging.INFO)
        if options.debug:
            self.logger.setLevel(level=logging.DEBUG)

        if options.rewind:
            # consumer.seek(pulsar.MessageId.earliest)
            self.logger.info('seeked to earliest message available as requested')

        self.setup()

        self.logger.info('start listening on topic %s', self.source.topic)
        # if batch_mode:
        #  self._run_batch()
        # else:

        if monitoring:
            self.monitor.expose()

        self.monitor.record_start()

        try:
            self._run_streaming()
            self.source.close()
            if not self.has_no_output():
                self.destination.close()
        except PipelineError as e:
            e.log(self.logger)
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.monitor.record_error(str(e))

        self.monitor.record_finish()



from .exception import PipelineError
from .message import Message
from .tap import DestinationTap, SourceTap, SourceOf, DestinationOf
from .worker import (
    GeneratorConfig, Generator,
    ProcessorConfig, Processor,
    SplitterConfig, Splitter,
)
from .cache import CacheOf
from .utils import parse_kind


__all__ = [
    'PipelineError',
    'SourceOf',
    'DestinationOf',
    'SourceTap',
    'DestinationTap',
    'GeneratorConfig',
    'Generator',
    'ProcessorConfig',
    'Processor',
    'SplitterConfig',
    'Splitter',
    'Message',
    'parse_kind',
    'CacheOf',
]



import argparse
import os

from .cache import KindsOfCache
from .tap import KindsOfSource


def parse_kind(args):
    kindParser = argparse.ArgumentParser(add_help=False)
    kindParser.add_argument('--kind', type=str, default=os.environ.get('PIPELINE', None),
                            choices=KindsOfSource(),
                            help='pipeline kind, can be {}'.format(','.join(KindsOfSource())))
    kindParser.add_argument('--cacheKind', type=str, default=os.environ.get('CACHEKIND', None),
                            choices=KindsOfCache(),
                            help='cache kind, can be {}'.format(','.join(KindsOfCache())))
    known, extras = kindParser.parse_known_args(args)
    return known, extras




from unittest import TestCase

from pipeline import Message


class TestMessage(TestCase):
    def test_default(self):
        m = Message()
        assert m.updated is False

    def test_get_version(self):
        m = Message()
        assert not m.get_version('tester')['version']

    def test_versions(self):
        m = Message()
        m.update_version('tester', [0, 1, 0])
        assert m.updated is True
        assert m.get_version('tester')['version'] == [0, 1, 0]
        assert m.get_version('tester')['order'] == 1
        m.updated = False
        m.update_version('tester', [0, 1, 1])
        assert m.updated is True
        assert m.get_version('tester')['version'] == [0, 1, 1]
        assert m.get_version('tester')['order'] == 1
        m.updated = False
        m.update_version('validator', [0, 0, 1])
        assert m.updated is True
        assert m.get_version('validator')['version'] == [0, 0, 1]
        assert m.get_version('validator')['order'] == 2

    def test_subclass(self):
        class NewMessage(Message):
            def __init__(self, other=None):
                super().__init__()
                if isinstance(other, dict):
                    self.dct = other
                else:
                    self.info, self.dct = other.info, other.dct

        newMessage1 = NewMessage({'key': 'new1'})
        newMessage2 = NewMessage(newMessage1)
        assert newMessage2.dct['key'] == 'new1'

    def test_update_replace(self):
        m = Message()
        m.update({'key': 'value'})
        assert m.get('key') == 'value'
        m.replace({'newkey': 'value'})
        assert m.get('key') is None
        assert m.get('newkey') == 'value'



import json
from abc import ABC
import time
import traceback

from .exception import PipelineError


class Message(ABC):
    keyname = 'key'

    def __init__(self, other=None):
        self.updated = False
        self.terminated = False
        self.info = {}
        self.dct = {}
        try:
            if other is not None:
                if isinstance(other, type(self)):
                    self.info = other.info
                    self.dct = other.dct
                elif isinstance(other, dict):
                    self.dct = other
                elif isinstance(other, bytes):
                    [self.info, self.dct] = self.deserialize(other)
                elif isinstance(other, str):
                    [self.info, self.dct] = self.deserialize(other)
                else:
                    raise PipelineError(
                        'Message needs to be initialized with a message, a dict or str/bytes, not "{}"'
                        .format(type(other)),
                        data=other
                    )
        except PipelineError as error:
            raise error
        except Exception as error:
            raise PipelineError(
                str(error), data=other, traceback=traceback.print_exception()
            )

    def __str__(self):
        return '{}<{}:{}>'.format(type(self).__name__, self.keyname, self.dct.get(self.keyname, None))

    def __repr__(self):
        return self.__str__()

    def __unicode__(self):
        return self.__str__()

    def log(self, logger):
        logger.warn(self.log_info)
        logger.warn(self.log_content)

    def log_info(self):
        return json.dumps(self.info, indent=4)

    def log_content(self):
        return json.dumps(self.dct, indent=4)

    @classmethod
    def add_arguments(cls, parser):
        return parser

    def serialize(self, indent=None):
        """serialize to binary."""
        return json.dumps([self.info, self.dct], indent=indent).encode('utf-8')

    @classmethod
    def deserialize(cls, raw):
        """deserialize to json."""
        if isinstance(raw, bytes):
            raw = raw.decode('utf-8')
        return json.loads(raw)

    def get_version(self, name):
        return self.info.setdefault(name, {'version': [], })

    def get_versions(self):
        return self.info

    # @abstractmethod
    # def publish_time(self):
    #  """ return publish_time """

    def is_valid(self):
        return True

    def should_update(self, name, version):
        versionDct = self.get_version(name)
        return version > versionDct['version']

    def update_version(self, name, version):
        versionDct = self.get_version(name)
        if version > versionDct['version']:
            versionDct['version'] = version
            versionDct['timestamp'] = time.time()
            if 'order' not in versionDct:
                versionDct['order'] = len(self.get_versions())
            self.updated = True

    def terminates(self):
        """ set terminated flag for this message """
        self.terminated = True

    def get(self, key, default=None):
        """ get value of key in content """
        return self.dct.get(key, default)

    def update(self, other):
        """ update message content """
        self.dct.update(other)

    def replace(self, other):
        """ replace message content """
        self.dct = other



import os
import tempfile
from unittest import TestCase

from pipeline import (
    Message,
    GeneratorConfig, Generator,
    ProcessorConfig, Processor,
    SplitterConfig, Splitter,
)


class TestWorkerCore(TestCase):
    def setUp(self):
        self.testDir = tempfile.TemporaryDirectory()
        self.infile = open(os.path.join(self.testDir.name, 'infile.txt'), 'w')
        self.infile.write("""[{},{"key":1, "language":"en"}]
                             [{},{"key":2, "language":"it"}]
                             [{},{"key":3, "language":"ar"}]""")
        self.infile.close()

    def test_mem_generator(self):
        class MyGenerator(Generator):
            def generate(self):
                for i in range(3):
                    yield {"key": i}
        generator = MyGenerator('generator', '0.1.0')
        generator.parse_args(args=['--kind', 'MEM', '--out-topic', 'test'])
        generator.start()
        assert len(generator.destination.results) == 3
        m = generator.destination.results[0]
        dct = m.get_version('generator')
        assert dct['version'] == [0, 1, 0]
        assert dct['order'] == 1

    def test_mem_generator_with_cache(self):
        class MyGenerator(Generator):
            def generate(self):
                for i in range(3):
                    self.cache.write(i, dict([(field, field) for field in self.cache.out_fields]))
                    yield {"key": i}
        memory = {}
        config = GeneratorConfig(cacheKind='MEM')
        generator = MyGenerator('generator', '0.1.0', config=config)
        generator.parse_args(
            args=['--kind', 'MEM', '--out-topic', 'test', '--out-fields', 'key1,key2'],
            config={'mem': memory}
        )
        generator.start()
        assert len(memory) == 3
        assert memory.get(1) is not None

    def test_generator_invalid_message(self):
        valids = [True, True, False]

        class InvalidMessage(Message):
            def is_valid(self):
                return self.dct.get('valid', False)

        class MyGenerator(Generator):
            def generate(self):
                for i, v in enumerate(valids):
                    yield {"key": i, "valid": v}

        config = GeneratorConfig(messageClass=InvalidMessage)
        generator = MyGenerator('generator', '0.1.0', config=config)
        generator.parse_args(args=['--kind', 'MEM', '--out-topic', 'test'])
        generator.start()
        assert len(generator.destination.results) == 2

    def test_processor_invalid_message(self):
        class InvalidMessage(Message):
            def is_valid(self):
                return False

        msgs = [{}, {}, {}]
        config = ProcessorConfig(messageClass=InvalidMessage)
        pro1 = Processor('tester1', '0.1.0', config=config)
        pro1.parse_args(args=['--kind', 'MEM', '--out-topic', 'test'], config={'data': msgs})
        pro1.use_retry_topic('optional-retry-topic')
        pro1.start()
        assert len(pro1.destination.results) == 0
        assert len(pro1.retryDestination.results) == 3

    def test_processor_retry(self):
        class RetryProcessor(Processor):
            def process(self, msg):
                return "Error"
        msgs = [{'key': '1'}, {'key': '2'}, {'key': '3'}]
        pro1 = RetryProcessor('tester1', '0.1.0')
        pro1.parse_args(args=['--kind', 'MEM', '--out-topic', 'test'], config={'data': msgs})
        pro1.use_retry_topic('optional-retry-topic')
        pro1.start()
        assert len(pro1.destination.results) == 0
        assert len(pro1.retryDestination.results) == 3

    def test_processor_terminated_messages(self):
        class RetryProcessor(Processor):
            def process(self, msg):
                msg.terminates()
                return None
        msgs = [{'key': '1'}, {'key': '2'}, {'key': '3'}]
        pro1 = RetryProcessor('tester1', '0.1.0')
        pro1.parse_args(args=['--kind', 'MEM', '--out-topic', 'test'], config={'data': msgs})
        pro1.use_retry_topic('optional-retry-topic')
        pro1.start()
        assert len(pro1.destination.results) == 0
        assert len(pro1.retryDestination.results) == 0

    def test_processor(self):
        class MyProcessor(Processor):
            def process(self, msg):
                updates = {'newkey': 'newval'}
                msg.update(updates)
                return None
        msgs = [{'key': '1'}, {'key': '2'}, {'key': '3'}]
        pro1 = MyProcessor('tester1', '0.1.0')
        pro1.parse_args(args=['--kind', 'MEM', '--out-topic', 'test'], config={'data': msgs})
        pro1.start()
        assert len(pro1.destination.results) == 3
        m = pro1.destination.results[0]
        assert m.get('newkey') == 'newval'

    def test_processor_with_cache(self):
        class MyProcessor(Processor):
            def process(self, msg):
                self.logger.info("%s, %s", msg.dct.get('key', None), msg.get('key'))
                self.cache.write(
                    msg.get('key'),
                    {'key3': self.cache.read(msg.get('key'))}
                )
                return None
        msgs = [{'key': 'm1'}, {'key': 'm2'}, {'key': 'm3'}]
        memory = {
            'm1': {'key1': 'val11', 'key2': 'val21'},
            'm2': {'key1': 'val12', 'key2': 'val22'},
            'm3': {'key1': 'val13', 'key2': 'val23'},
        }
        config = ProcessorConfig(cacheKind='MEM')
        pro1 = MyProcessor('tester1', '0.1.0', config=config)
        pro1.parse_args(
            args='--kind MEM --out-topic test --in-fields key1,key2 --out-fields key3'.split(),
            config={
                'data': msgs,
                'mem': memory,
            }
        )
        pro1.start()
        assert len(pro1.destination.results) == 3
        m = memory.get('m1')
        assert m.get('key3')['key1'] == 'val11'
        assert m.get('key3')['key2'] == 'val21'

    def test_file(self):
        processor = Processor('fileprocessor', '0.1.0')
        processor.parse_args(args=['--kind', 'FILE', '--in-topic', 'test', '--out-topic', 'test',
                                   '--infile', os.path.join(self.testDir.name, 'infile.txt'),
                                   '--outfile', os.path.join(self.testDir.name, 'outfile.txt')])
        processor.start()
        with open(os.path.join(self.testDir.name, 'outfile.txt'), 'r') as f:
            assert len(list(f)) == 3

    def test_file_stdin_stdout(self):
        processor = Processor('fileprocessor', '0.1.0')
        processor.parse_args(args=['--kind', 'FILE', '--in-topic', 'test', '--out-topic', 'test',
                                   '--infile', os.path.join(self.testDir.name, 'infile.txt'),
                                   '--outfile', '-'])
        processor.start()

    def test_file_repeat(self):
        processor = Processor('fileprocessor', '0.1.0')
        processor.parse_args(args=['--kind', 'FILE', '--in-topic', 'test', '--out-topic', 'test',
                                   '--infile', os.path.join(self.testDir.name, 'infile.txt'),
                                   '--repeat', "2",
                                   '--outfile', os.path.join(self.testDir.name, 'outfile.txt')])
        processor.start()
        with open(os.path.join(self.testDir.name, 'outfile.txt'), 'r') as f:
            assert len(list(f)) == 6

    def test_mem_processor(self):
        msgs = [{}, {}, {}]
        pro1 = Processor('tester1', '0.1.0')
        pro1.parse_args(args=['--kind', 'MEM', '--out-topic', 'test'], config={'data': msgs})
        pro1.start()
        assert len(pro1.destination.results) == 3

    def test_mem_processor_nooutput(self):
        config = ProcessorConfig(noOutput=True)
        pro1 = Processor('tester2', '0.1.0', config=config)
        pro1.parse_args(args=['--kind', 'MEM'], config={'data': [{}]})
        pro1.start()
        assert not hasattr(pro1, 'destination')

    def test_splitter(self):
        msgs = [{'language': 'en'}, {'language': 'it'}]
        splitter = Splitter('spliter1', '0.1.0')
        splitter.parse_args(args=['--kind', 'MEM', '--out-topic', 'test'], config={'data': msgs})
        splitter.start()
        assert len(splitter.destinations['test-en'].results) == 1
        assert splitter.destinations['test-en'].results[0].dct['language'] == 'en'
        assert len(splitter.destinations['test-it'].results) == 1
        assert splitter.destinations['test-it'].results[0].dct['language'] == 'it'

    def test_splitter_invalid_message(self):
        class InvalidMessage(Message):
            def is_valid(self):
                return False

        msgs = [{"key": 1, 'language': 'en'}, {"key": 2, 'language': 'it'}]
        config = SplitterConfig(messageClass=InvalidMessage)
        splitter = Splitter('spliter1', '0.1.0', config=config)
        splitter.parse_args(args=['--kind', 'MEM', '--out-topic', 'test'], config={'data': msgs})
        splitter.start()
        assert len(splitter.destinations['test-en'].results) == 0
        assert len(splitter.destinations['test-it'].results) == 0

    def test_splitter_file(self):
        splitter = Splitter('spliter1', '0.1.0')
        splitter.parse_args(args=['--kind', 'FILE', '--in-topic', 'test', '--out-topic', 'test',
                                  '--infile', os.path.join(self.testDir.name, 'infile.txt'),
                                  '--outfile', os.path.join(self.testDir.name, 'outfile.txt')])
        splitter.start()
        assert len(splitter.destinations) == 3

    def test_custom_message(self):
        class CustomMessage(Message):
            def __str__(self):
                return 'Message({}: {})'.format(self.dct['key'], self.dct['value'])
        config = ProcessorConfig(messageClass=CustomMessage)
        pro1 = Processor('tester3', '0.1.0', config=config)
        pro1.parse_args(args=['--kind', 'MEM'], config={'data': [{'key': 'key1', 'value': 'value1'}]})
        pro1.start()
        assert pro1.destination.results[0].get('key') == 'key1'



#!/usr/bin/env python
import json
import logging
import os
import sys
from abc import ABC, abstractmethod

import pulsar
from confluent_kafka import (
    OFFSET_BEGINNING, Consumer, KafkaError,
    KafkaException, Producer
)

from .message import Message

FORMAT = '%(asctime)-15s %(levelname)s %(message)s'
logging.basicConfig(format=FORMAT)
logger = logging.getLogger('worker')
logger.setLevel(logging.DEBUG)


class SourceTap(ABC):
    """ Tap defines the interface for connecting components in pipeline.
    A source will emit Message.
    """
    kind = 'NONE'

    def __init__(self, config, logger=logger):
        self.config = config
        self.rewind = config.rewind
        self.topic = config.in_topic
        self.logger = logger
        self.messageClass = config.message if hasattr(config, 'message') else Message

    @abstractmethod
    def read(self):
        """receive message."""

    def rewind(self):
        """rewind to earliest message."""
        pass

    @classmethod
    def add_arguments(cls, parser):
        parser.add_argument('--rewind', action='store_true', default=False,
                            help='read from earliest message')
        parser.add_argument('--in-topic', type=str,
                            default=os.environ.get('INTOPIC', 'in-topic'),
                            help='pulsar topic to read, exclude tenant and namespace')

    @classmethod
    def is_cls_of(cls, kind):
        return kind == cls.kind

    def close(self):
        pass

    def acknowledge(self):
        pass


class DestinationTap(ABC):
    """ Tap defines the interface for connecting components in pipeline.
    """
    kind = 'NONE'

    def __init__(self, config, logger=logger):
        self.config = config
        self.topic = config.out_topic
        self.logger = logger
        self.messageClass = config.message if hasattr(config, 'message') else Message

    @abstractmethod
    def write(self, message):
        """send message."""

    @classmethod
    def is_cls_of(cls, kind):
        return kind == cls.kind

    @classmethod
    def add_arguments(cls, parser):
        parser.add_argument('--out-topic', type=str, default=os.environ.get('OUTTOPIC', 'out-topic'),
                            help='topic to write, for pulsar, exclude tenant and namespace')

    def close(self):
        pass


def KindsOfSource():
    return [cls.kind for cls in SourceTap.__subclasses__()]


def SourceOf(typename):
    for cls in SourceTap.__subclasses__():
        if cls.is_cls_of(typename):
            return cls


def DestinationOf(typename):
    for cls in DestinationTap.__subclasses__():
        if cls.is_cls_of(typename):
            return cls


class MemorySource(SourceTap):
    """ MemorySource iterates over a list of dict from 'data' in config.
    It is for testing only.

    >>> from types import SimpleNamespace
    >>> data = [{'id':1},{'id':2}]
    >>> [ m.dct for m in MemorySource(SimpleNamespace(rewind=False, in_topic='test', data=data)).read() ]
    [{'id': 1}, {'id': 2}]
    """
    kind = 'MEM'

    def read(self):
        for i in self.config.data:
            yield self.messageClass(i)

    def rewind(self):
        pass


class MemoryDestination(DestinationTap):
    """ MemoryDestination stores dicts written in results.
    It is for testing only.

    >>> from types import SimpleNamespace
    >>> d = MemoryDestination(SimpleNamespace(out_topic='test'))
    >>> d.write(Message({"id": 1}))
    >>> d.write(Message({"id": 2}))
    >>> [r.dct for r in d.results]
    [{'id': 1}, {'id': 2}]
    """
    kind = 'MEM'

    def __init__(self, config, logger=logger):
        super().__init__(config, logger)
        self.results = []

    def write(self, message):
        self.results.append(message)


class FileSource(SourceTap):
    """ FileSource iterates over lines from a input
    text file (utf-8), each line should be a json string for a dict.
    It can be used for integration test for workers.

    >>> import tempfile
    >>> from argparse import ArgumentParser
    >>> parser = ArgumentParser()
    >>> FileSource.add_arguments(parser)
    >>> with tempfile.NamedTemporaryFile() as tmpfile:
    ...     tmpfile.write(b'[{ }, {"id": 0}]') and True
    ...     tmpfile.flush()
    ...     config = parser.parse_args("--infile {}".format(tmpfile.name).split())
    ...     fileSource = FileSource(config)
    ...     [m.dct["id"] for m in fileSource.read()]
    True
    [0]
    """
    kind = 'FILE'

    def __init__(self, config, logger=logger):
        super().__init__(config, logger)
        if config.infile == '-':
            self.infile = sys.stdin
        else:
            self.infile = open(config.infile, 'r')
        self.repeat = config.repeat
        self.logger.info('File Source: %s (repeat %d)', config.infile, config.repeat)

    def __repr__(self):
        return 'FileSource("{}")'.format(self.infile.name)

    @classmethod
    def add_arguments(cls, parser):
        super().add_arguments(parser)
        parser.add_argument('--infile', type=str, required=True,
                            help='input file containing one message in JSON format per line')
        parser.add_argument('--repeat', type=int, default=1,
                            help='repeat input N times')

    def read(self):
        for i in range(0, self.repeat):
            for line in self.infile:
                yield self.messageClass(line)
            if self.repeat > 1:
                self.infile.seek(0, 0)


class FileDestination(DestinationTap):
    """ FileDestination writes items to an output file, one item per line in json format.

    >>> import os, tempfile
    >>> from argparse import ArgumentParser
    >>> parser = ArgumentParser()
    >>> FileDestination.add_arguments(parser)
    >>> tmpdir = tempfile.mkdtemp()
    >>> outFilename = os.path.join(tmpdir, 'outfile.txt')
    >>> config = parser.parse_args("--outfile {}".format(outFilename).split())
    >>> FileDestination(config)
    FileDestination("...outfile.txt")
    """
    kind = 'FILE'

    def __init__(self, config, logger=logger):
        super().__init__(config, logger)
        self.filename = config.outfile
        if config.outfile == '-':
            self.outFile = sys.stdout
        else:
            self.outFile = open(config.outfile, 'w')
        self.logger.info('File Destination: %s', config.outfile)

    def __repr__(self):
        return 'FileDestination("{}")'.format(self.outFile.name)

    @classmethod
    def add_arguments(cls, parser):
        super().add_arguments(parser)
        parser.add_argument('--outfile', type=str, required=True,
                            help='output json file')

    def write(self, message):
        print(message.serialize().decode('utf-8'), file=self.outFile)

    def close(self):
        if self.filename != '-':
            self.outFile.close()
            self.logger.info('File Destination closed')


class KafkaSource(SourceTap):
    """ KafkaSource reads from KAFKA

    >>> from argparse import ArgumentParser
    >>> parser = ArgumentParser()
    >>> KafkaSource.add_arguments(parser)
    >>> config = parser.parse_args(["--config", '{"sasl.mechanisms": "PLAIN"}'])
    >>> KafkaSource(config)
    KafkaSource(host="kafka.kafka.svc.cluster.local",groupid="group-id",topic="in-topic")
    """
    kind = 'KAFKA'

    def __init__(self, config, logger=logger):
        super().__init__(config, logger)

        kafkaConfig = {
            'bootstrap.servers': config.kafka,
            'broker.address.family': 'v4',
            'group.id': config.group_id,
            'auto.offset.reset': 'earliest',
            # 'log_level': 0,
            # 'debug': 'consumer',
            'enable.auto.commit': 'false',
            'session.timeout.ms': config.timeout
        }

        if config.config:
            extraConfig = json.loads(config.config)
            kafkaConfig.update(extraConfig)

        self.consumer = Consumer(kafkaConfig, logger=self.logger)
        self.topic = config.in_topic
        self.last_msg = None

        def maybe_rewind(c, partitions):
            self.logger.info('Assignment: %s', str(partitions))
            if config.rewind:
                for partition in partitions:
                    partition.offset = OFFSET_BEGINNING
                self.logger.info('Rewind, new assignment: %s', str(partitions))
                c.assign(partitions)

        self.consumer.subscribe([self.topic], on_assign=maybe_rewind)
        self.logger.info('KAFKA consumer subscribed to topic %s', self.topic)

    def __repr__(self):
        return 'KafkaSource(host="{}",groupid="{}",topic="{}")'.format(
            self.config.kafka,
            self.config.group_id,
            self.topic
        )

    @classmethod
    def add_arguments(cls, parser):
        super().add_arguments(parser)
        parser.add_argument('--kafka', type=str,
                            default=os.environ.get('KAFKA', 'kafka.kafka.svc.cluster.local'),
                            help='kafka address')
        parser.add_argument('--group-id', type=str,
                            default=os.environ.get('GROUPID', 'group-id'),
                            help='group id')
        parser.add_argument('--timeout', type=int,
                            default=os.environ.get('TIMEOUT', 30000),
                            help='request timeout')
        parser.add_argument('--config', type=str,
                            default=os.environ.get('KAFKACONFIG', None),
                            help='kafka configuration in JSON format')
        parser.add_argument('--poll-timeout', type=int, default=30,
                            help='poll new message timeout in seconds')

    def read(self):
        while True:
            msg = self.consumer.poll(timeout=self.config.poll_timeout)
            if msg is None:
                self.logger.warn('No message to read, timed out')
                break

            if msg.error():
                if msg.error().code() == KafkaError._PARTITION_EOF:
                    self.logger.warn('Reaching EOF')
                    break
                else:
                    raise KafkaException(msg.error())
            else:
                self.logger.info('Read {}, {}'.format(msg.topic(), msg.offset()))
                self.last_msg = msg
                yield self.messageClass(msg.value())

    def acknowledge(self):
        if self.last_msg:
            self.consumer.commit(message=self.last_msg)

    def close(self):
        self.consumer.close()


class KafkaDestination(DestinationTap):
    """ KafkaDestination writes to KAFKA

    >>> from argparse import ArgumentParser
    >>> parser = ArgumentParser()
    >>> KafkaDestination.add_arguments(parser)
    >>> config = parser.parse_args(["--config", '{"sasl.mechanisms": "PLAIN"}'])
    >>> KafkaDestination(config)
    KafkaDestination(host="kafka.kafka.svc.cluster.local",topic="out-topic")
    """
    kind = 'KAFKA'

    def __init__(self, config, logger=logger):
        super().__init__(config, logger)
        kafkaConfig = {
            'bootstrap.servers': config.kafka,
            'queue.buffering.max.ms': 100,
            'message.send.max.retries': 5,
            'request.required.acks': 'all',
            'broker.address.family': 'v4',
        }

        if config.config:
            extraConfig = json.loads(config.config)
            kafkaConfig.update(extraConfig)

        self.topic = config.out_topic
        self.producer = Producer(kafkaConfig, logger=self.logger)

    def __repr__(self):
        return 'KafkaDestination(host="{}",topic="{}")'.format(
            self.config.kafka,
            self.topic
        )

    @classmethod
    def add_arguments(cls, parser):
        super().add_arguments(parser)
        parser.add_argument('--kafka', type=str,
                            default=os.environ.get('KAFKA', 'kafka.kafka.svc.cluster.local'),
                            help='kafka address')
        parser.add_argument('--config', type=str,
                            default=os.environ.get('KAFKACONFIG', None),
                            help='kafka configuration in JSON format')

    def write(self, message):
        def delivery_report(err, msg):
            if err:
                self.logger.error('Message delivery failed ({} [{}]: {}'.format(
                    msg.topic(),
                    str(msg.partition()),
                    err
                ))
            else:
                self.logger.info('Message delivered to {} [{}] {}'.format(
                    msg.topic(),
                    msg.partition(),
                    msg.offset()
                ))
        self.producer.produce(self.topic, message.serialize(), callback=delivery_report)
        self.producer.flush()

    def close(self):
        self.producer.flush()


class PulsarSource(SourceTap):
    """ PulsarSource reads from Pulsar

    >>> from unittest.mock import patch
    >>> from argparse import ArgumentParser
    >>> parser = ArgumentParser()
    >>> PulsarSource.add_arguments(parser)
    >>> config = parser.parse_args([])
    >>> with patch('pulsar.Client') as c:
    ...     PulsarSource(config)
    PulsarSource(host="pulsar://pulsar.pulsar.svc.cluster.local:6650",name="persistent://meganews/test/in-topic",subscription="subscription")
    """
    kind = 'PULSAR'

    def __init__(self, config, logger=logger):
        super().__init__(config, logger)
        self.config = config
        self.client = pulsar.Client(config.pulsar)
        self.tenant = config.tenant
        self.namespace = config.namespace
        self.topic = config.in_topic
        self.subscription = config.subscription
        self.name = 'persistent://{}/{}/{}'.format(
            self.tenant,
            self.namespace,
            self.topic
        )
        self.consumer = self.client.subscribe(
            self.name,
            self.subscription,
            receiver_queue_size=1,
            consumer_type=pulsar.ConsumerType.Shared,
        )
        self.last_msg = None

    def __repr__(self):
        return 'PulsarSource(host="{}",name="{}",subscription="{}")'.format(
            self.config.pulsar,
            self.name,
            self.subscription,
        )

    @classmethod
    def add_arguments(cls, parser):
        super().add_arguments(parser)
        parser.add_argument('--pulsar', type=str,
                            default=os.environ.get('PULSAR', 'pulsar://pulsar.pulsar.svc.cluster.local:6650'),
                            help='pulsar address')
        parser.add_argument('--tenant', type=str,
                            default=os.environ.get('TENANT', 'meganews'),
                            help='pulsar tenant, always is meganews')
        parser.add_argument('--namespace', type=str,
                            default=os.environ.get('NAMESPACE', 'test'),
                            help='pulsar namespace, by default test')
        parser.add_argument('--subscription', type=str,
                            default=os.environ.get('SUBSCRIPTION', 'subscription'),
                            help='subscription to read')

    def read(self, timeout=0):
        while True:
            try:
                msg = self.consumer.receive()
                self.last_msg = msg
                yield self.messageClass(msg.data())
            except Exception as ex:
                self.logger.error(ex)
                break

    def acknowledge(self):
        self.consumer.acknowledge(self.last_msg)

    def close(self):
        self.client.close()


class PulsarDestination(DestinationTap):
    """ PulsarDestination writes to Pulsar

    >>> from unittest.mock import patch
    >>> from argparse import ArgumentParser
    >>> parser = ArgumentParser()
    >>> PulsarDestination.add_arguments(parser)
    >>> config = parser.parse_args([])
    >>> with patch('pulsar.Client') as c:
    ...     PulsarDestination(config)
    PulsarDestination(host="pulsar://pulsar.pulsar.svc.cluster.local:6650",name="persistent://meganews/test/out-topic")
    """
    kind = 'PULSAR'

    def __init__(self, config, logger=logger):
        super().__init__(config, logger)
        self.config = config
        self.client = pulsar.Client(config.pulsar)
        self.tenant = config.tenant
        self.namespace = config.namespace
        self.topic = config.out_topic
        self.name = 'persistent://{}/{}/{}'.format(
            self.tenant, self.namespace, self.topic)
        self.producer = self.client.create_producer(self.name)

    def __repr__(self):
        return 'PulsarDestination(host="{}",name="{}")'.format(
            self.config.pulsar,
            self.name,
        )

    @classmethod
    def add_arguments(cls, parser):
        super().add_arguments(parser)
        parser.add_argument('--pulsar', type=str,
                            default=os.environ.get('PULSAR', 'pulsar://pulsar.pulsar.svc.cluster.local:6650'),
                            help='pulsar address')
        parser.add_argument('--tenant', type=str,
                            default=os.environ.get('TENANT', 'meganews'),
                            help='pulsar tenant, always is meganews')
        parser.add_argument('--namespace', type=str,
                            default=os.environ.get('NAMESPACE', 'test'),
                            help='pulsar namespace, by default test')

    def write(self, message):
        self.producer.send(message.serialize())

    def close(self):
        self.client.close()



#!/usr/bin/env python
import codecs
import os

from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), "requirements.txt")) as f:
    required = f.read().splitlines()

with codecs.open(os.path.join(os.path.dirname(__file__), "README.rst"), "r", "utf-8") as f:
    readme = f.read()

setup(name="tanbih-pipeline",
      version="0.4.2",
      description="a pipeline framework for streaming processing",
      long_description=readme,
      long_description_content_type='text/x-rst',
      author="yifan",
      author_email="yzhang@hbku.edu.qa",
      url="https://github.com/yifan/pipeline",
      license="MIT",
      classifiers=[
          "License :: OSI Approved :: MIT License",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.7",
      ],
      packages=find_packages("./src"),
      package_dir={"": "src"},
      py_modules=["pipeline", ],
      include_package_data=True,
      zip_safe=False,
      install_requires=required,
      )



import logging
import time

from prometheus_client import Counter, Enum, Info, start_http_server

FORMAT = '%(asctime)-15s %(levelname)s %(message)s'
logging.basicConfig(format=FORMAT)
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class Monitor(object):
    info = Info('worker', 'worker information')
    err = Info('worker_error', 'worker error')
    counter = Counter(
        'worker_operation',
        'counter for worker operations',
        ['name', 'operation', 'topic']
    )
    state = Enum(
        'worker_state',
        'state of worker',
        ['name'],
        states=['starting', 'running', 'stopped'],
    )

    def __init__(self, worker, port=8000):
        self.worker = worker
        self.state.labels(name=worker.name).state('starting')
        self.port = port

    def expose(self):
        start_http_server(self.port)

    def record_worker_info(self):
        self.info.info({
            'name': self.worker.name,
            'version': self.worker.version,
            'description': self.worker.description,
            'in-topic': None if self.worker.has_no_input() else self.worker.source.topic,
            'out-topic': None if self.worker.has_no_output() else self.worker.destination.topic,
        })

    def record_start(self):
        self.info.info({
            'name': self.worker.name,
            'time': str(time.time()),
            'event': 'start',
        })
        self.state.labels(name=self.worker.name).state('running')

    def record_finish(self):
        self.info.info({
            'name': self.worker.name,
            'time': str(time.time()),
            'event': 'finish',
        })
        self.state.labels(name=self.worker.name).state('stopped')

    def record_error(self, msg):
        self.err.info({
            'name': self.worker.name,
            'time': str(time.time()),
            'event': 'error',
            'details': msg,
        })
        self.counter.labels(name=self.worker.name, operation='error', topic=None).inc()

    def record_write(self, topic=None):
        self.counter.labels(name=self.worker.name, operation='write', topic=topic).inc()

    def record_read(self, topic=None):
        self.counter.labels(name=self.worker.name, operation='read', topic=topic).inc()
